/*
 * Copyright 2010 Boris Kochergin. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef GEOM_HPP
#define GEOM_HPP

#include <string>
#include <vector>
#include <algorithm>

#include <libgeom.h>

class Geom {
  public:
    Geom(const std::string &className, const std::string &geomName);
    std::string name() const;
    const std::string &error() const;
    bool update();
    const bool &lastUpdate() const;
    const std::string &status() const;
    const std::string &lastStatus() const;
    const std::vector <std::string> &providers() const;
    const std::vector <std::string> &lastProviders() const;
  private:
    std::string _className;
    std::string _geomName;
    gmesh gMesh;
    gclass *gClass;
    ggeom *gGeom;
    gconfig *gConf;
    gconsumer *gConsumer;
    gprovider *gProvider;
    std::string _status;
    std::string _lastStatus;
    std::vector <std::string> _providers;
    std::vector <std::string> _lastProviders;
    bool _lastUpdate;
    int ret;
    std::string errorMessage;
};

Geom::Geom(const std::string &className, const std::string &geomName) {
  _lastUpdate = true;
  _className = className;
  _geomName = geomName;
}

std::string Geom::name() const {
  return _className + '/' + _geomName;
}

const std::string &Geom::error() const {
  return errorMessage;
}

bool Geom::update() {
  ret = geom_gettree(&gMesh);
  if (ret != 0) {
    _lastUpdate = false;
    errorMessage = "geom_gettree(): ";
    errorMessage += strerror(ret);
    return false;
  }
  gClass = NULL;
  LIST_FOREACH(gClass, &(gMesh.lg_class), lg_class) {
    if (gClass -> lg_name == _className) {
      break;
    }
  }
  if (gClass == NULL) {
    _lastUpdate = false;
    errorMessage = "update(): " + _className + ": class not found";
    return false;
  }
  gGeom = NULL;
  LIST_FOREACH(gGeom, &(gClass -> lg_geom), lg_geom) {
    if (gGeom -> lg_name == _geomName) {
      break;
    }
  }
  if (gGeom == NULL) {
    _lastUpdate = false;
    errorMessage = "update(): " + _className + '/' + _geomName +
                   ": geom not found";
    return false;
  }
  LIST_FOREACH(gConf, &(gGeom -> lg_config), lg_config) {
    if (strcasecmp(gConf -> lg_name, "state") == 0) {
      break;
    }
  }
  _lastStatus = _status;
  if (gConf == NULL) {
    _status = "N/A";
  }
  else {
    _status = gConf -> lg_val;
  }
  _lastProviders = _providers;
  _providers.clear();
  LIST_FOREACH(gConsumer, &(gGeom -> lg_consumer), lg_consumer) {
    gProvider = gConsumer -> lg_provider;
    if (gProvider == NULL) {
      _lastUpdate = false;
      errorMessage = "update(): " + _className + '/' + _geomName +
                     ": no providers found";
      return false;
    }
    _providers.push_back(gProvider -> lg_name);
  }
  sort(_providers.begin(), _providers.end());
  _lastUpdate = true;
  return true;
}

const bool &Geom::lastUpdate() const {
  return _lastUpdate;
}

const std::string &Geom::status() const {
 return _status;
}

const std::string &Geom::lastStatus() const {
 return _lastStatus;
}

const std::vector <std::string> &Geom::providers() const {
  return _providers;
}

const std::vector <std::string> &Geom::lastProviders() const {
  return _lastProviders;
}

#endif
