/*
 * Copyright 2009-2010 Boris Kochergin. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>
#include <algorithm>

#include <signal.h>
#include <sys/types.h>
#include <unistd.h>

#include "configuration.hpp"
#include "geom.hpp"
#include "smtp.hpp"

using namespace std;

string pidFileName = "/var/run/geomWatch.pid";

void signalHandler(int signal) {
  switch (signal) {
    case SIGTERM:
      unlink(pidFileName.c_str());
      exit(0);
  }
}

int main(int argc, char *argv[]) {
  char option;
  ofstream pidFile;
  Configuration conf;
  vector <Geom> geoms;
  SMTP smtp;
  string geom, configFileName = "/usr/local/etc/geomWatch.conf";
  size_t slash;
  bool lastUpdate;
  long hostnameLength = sysconf(_SC_HOST_NAME_MAX);
  char *hostname;
  ostringstream message;
  vector <string> lost;
  pid_t pid;
  signal(SIGTERM, signalHandler);
  while ((option = getopt(argc, argv, "c:p:")) != -1) {
    switch (option) {
      case 'c':
        configFileName = optarg;
        break;
      case 'p':
        pidFileName = optarg;
        break;
      default:
        return 1;
    }
  }
  if (!conf.open(configFileName)) {
    cerr << argv[0] << ": " << conf.error() << endl;
    return 1;
  }
  if (conf.getStrings("geom").size() == 0) {
    cerr << argv[0] << ": no geoms specified" << endl;
    return 1;
  }
  if (conf.getStrings("recipient").size() == 0) {
    cerr << argv[0] << ": no recipients specified" << endl;
    return 1;
  }
  if (conf.getNumber("interval") == std::numeric_limits <size_t>::max()) {
    cerr << argv[0] << ": no interval specified" << endl;
    return 1;
  }
  if (conf.getNumber("smtpServer") == std::numeric_limits <size_t>::max()) {
    cerr << argv[0] << ": no SMTP server specified" << endl;
    return 1;
  }
  if (hostnameLength == -1) {
    cerr << argv[0] << ": sysconf(_SC_HOST_NAME_MAX): " << strerror(errno)
         << endl;
    return 1;
  }
  hostname = (char*)malloc(hostnameLength + 1);
  if (hostname == NULL) {
    cerr << argv[0] << ": malloc(" << hostnameLength << "): " << strerror(errno)
         << endl;
    return 1;
  }
  if (gethostname(hostname, hostnameLength + 1) == -1) {
    cerr << argv[0] << ": gethostname(): " << strerror(errno) << endl;
    return 1;
  }
  for (size_t i = 0; i < conf.getStrings("geom").size(); ++i) {
    geom = conf.getStrings("geom")[i];
    slash = geom.find('/');
    if (slash == string::npos) {
      cerr << argv[0] << ": no geom specified for class " << geom << endl;
      return 1;
    }
    geoms.push_back(Geom(geom.substr(0, slash), geom.substr(slash + 1)));
    if (!(geoms.rbegin() -> update())) {
      cerr << argv[0] << ": " << geoms.rbegin() -> error() << endl;
      return 1;
    }
  }
  if (!smtp.initialize(conf.getString("smtpServer"), conf.getNumber("smtpAuth"),
                       conf.getString("smtpUser"),
                       conf.getString("smtpPassword"),
                       conf.getStrings("recipient"))) {
    cerr << argv[0] << ": " << smtp.error() << endl;
    return 1;
  }
  if (!smtp.setFrom(conf.getString("from"))) {
    cerr << argv[0] << ": " << smtp.error() << endl;
    return 1;
  }
  pid = fork();
  if (pid < 0) {
    cerr << argv[0] << ": fork(): " << strerror(errno) << endl;
    return 1;
  }
  /* Exit if this is the parent process. */
  if (pid != 0) {
    return 0;
  }
  /*
   * Otherwise, write the daemonized child's PID to the PID file and enter the
   * main GEOM-checking loop.
   */
  pidFile.open(pidFileName.c_str());
  if (!pidFile) {
    cerr << argv[0] << ": open(): " << pidFileName << ": " << strerror(errno)
         << endl;
    return 1;
  }
  pidFile << getpid() << endl;
  pidFile.close();
  while (1) {
    sleep(conf.getNumber("interval"));
    for (size_t i = 0; i < geoms.size(); ++i) {
      lastUpdate = geoms[i].lastUpdate();
      if (!(geoms[i].update())) {
        smtp.setSubject("Problem checking " + geoms[i].name() + " on " +
                        hostname);
        if (lastUpdate == true) {
          smtp.setSubject("Problem checking " + geoms[i].name() + " on " +
                          hostname);
          message << "The following problem was encountered while checking " 
                  << "the status of " << geoms[i].name() << ": " << endl
                  << endl << '\t' << geoms[i].error() << endl;
          smtp.setMessage(message.str());
          message.str("");
          if (!smtp.send()) {
            cerr << argv[0] << ": " << smtp.error() << endl;
          }
        }
      }
      else {
        if (geoms[i].providers().size() < geoms[i].lastProviders().size()) {
          smtp.setSubject("Component lost for " + geoms[i].name() + " on " +
                          hostname);
          if (geoms[i].status() == geoms[i].lastStatus()) {
            message << "The status of " << geoms[i].name() << " remains "
                    << "unchanged from " << geoms[i].status();
          }
          else {
            message << "The status of " << geoms[i].name() << " has changed "
                    << "from " << geoms[i].lastStatus() << " to "
                    << geoms[i].status();
          }
          lost.resize(geoms[i].lastProviders().size() -
                      geoms[i].providers().size());
          set_difference(geoms[i].lastProviders().begin(),
                         geoms[i].lastProviders().end(),
                         geoms[i].providers().begin(),
                         geoms[i].providers().end(),
                         lost.begin());
          message << ". The following components have been lost:" << endl
                  << endl;
          for (size_t j = 0; j < lost.size(); ++j) {
            message << '\t' << lost[j] << endl;
          }
          message << endl << "Remaining components:" << endl << endl;
          for (size_t j = 0; j < geoms[i].providers().size(); ++j) {
            message << '\t' << geoms[i].providers()[j] << endl;
          }
          smtp.setMessage(message.str());
          message.str("");
          if (!smtp.send()) {
            cerr << argv[0] << ": " << smtp.error() << endl;
          }
        }
      }
    }
  }
  /* Not reached. */
  return 0;
}
